//----------------------------------------------------------------------------------
// Calculate volume
// 2007 Dennis Hawthorne
//----------------------------------------------------------------------------------
#ifdef _WINDOWS
#ifndef WINDOWS
#define WINDOWS
#endif
#endif

#include "ac_plugin.h"
#include <math.h>
#include <stdio.h>
//----------------------------------------------------------------------------------
static void ddh_calc_volume()
{

   // Find unclosed geometry
   bool open_geometry = false;
   bool complex_geometry = false;
   List *edgelist = ac_selection_get_edges( true );
   for( List *pe = edgelist; pe != NULL; pe = pe->next )
   {
      ACEdge *e = ( ACEdge * )( pe->data );
      if( e->connected == 1 )
      {
         open_geometry = true;
      }
      if( e->connected > 2 )
      {
         complex_geometry = true;
      }
      if( complex_geometry && open_geometry )
      {
         break;
      }
   }
   ac_edgelist_free( &edgelist );

   // Get triangulated selection
   List *seltris = NULL;
   List *surfall = ac_selection_get_whole_surfaces_all();
   for( List *ps = surfall; ps != NULL; ps = ps->next )
   {
      List *pts = surface_get_triangulations(( Surface * )( ps->data ));
      for( List *ppts = pts; pts != NULL; pts = pts->next )
      {
         list_add_item_head( &seltris, ppts->data );
      }
      list_free( &pts );
   }
   list_free( &surfall );

   // Get the X-offset to start all X calcs from 0
   float minx;
   bool first = true;
   for( List *ps = seltris; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      Vertex *v1 = (( SVertex * )( s->vertlist->data ))->v;
      Vertex *v2 = (( SVertex * )( s->vertlist->next->data ))->v;
      Vertex *v3 = (( SVertex * )( s->vertlist->next->next->data ))->v;
      if( first )
      {
         minx = min( min( v1->x, v2->x ), v3->x );
         first = false;
      }
      else
      {
         minx = min( minx, min( min( v1->x, v2->x ), v3->x ));
      }
   }

   // Get volume using divergence theorem calculation
   /**/
   double totalvol = 0;
   for( List *ps = seltris; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      Vertex *v1 = (( SVertex * )( s->vertlist->data ))->v;
      Vertex *v2 = (( SVertex * )( s->vertlist->next->data ))->v;
      Vertex *v3 = (( SVertex * )( s->vertlist->next->next->data ))->v;
      totalvol += ((((( double )( v2->y )) - (( double )( v1->y ))) * ((( double )( v3->z )) - (( double )( v1->z )))) -
                   (((( double )( v2->z )) - (( double )( v1->z ))) * ((( double )( v3->y )) - (( double )( v1->y ))))) *
                  ((( double )( v1->x - minx )) + (( double )( v2->x - minx )) + (( double )( v3->x - minx )));
      /*
      printf( "---------------------------------------------------------------\n" );
      printf( "v1,%g,%g,%g\n", v1->x, v1->y, v1->z );
      printf( "v2,%g,%g,%g\n", v2->x, v2->y, v2->z );
      printf( "v3,%g,%g,%g\n", v3->x, v3->y, v3->z );
      printf( "minx,%g\n", minx );
      printf( "x,%g,%g,%g\n", (( double )( v1->x )), (( double )( v2->x )), (( double )( v3->x )));
      printf( "vol,%g\n", ((((( double )( v2->y )) - (( double )( v1->y ))) * ((( double )( v3->z )) - (( double )( v1->z )))) -
                   (((( double )( v2->z )) - (( double )( v1->z ))) * ((( double )( v3->y )) - (( double )( v1->y ))))) *
                  ((( double )( v1->x - minx )) + (( double )( v2->x - minx )) + (( double )( v3->x - minx ))));
      /**/
   }
   totalvol /= 6.0;
   printf( "Total volume = %g\n", totalvol );
   /**/

   /**
   double totalarea = 0;
   double totalvol = 0;
   struct DVertexS { double x,y,z; };
   for( List *ps = seltris; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      Vertex *v1 = (( SVertex * )( s->vertlist->data ))->v;
      Vertex *v2 = (( SVertex * )( s->vertlist->next->data ))->v;
      Vertex *v3 = (( SVertex * )( s->vertlist->next->next->data ))->v;
      DVertexS va, vb, vc;
      // Length of 2 sides of tri
      va.x = v2->x;
      va.y = v2->y;
      va.z = v2->z;
      vb.x = v3->x;
      vb.y = v3->y;
      vb.z = v3->z;
      //memcpy( &va, v2, sizeof( Vertex ));
      //memcpy( &vb, v3, sizeof( Vertex ));
      va.x -= ( double )( v1->x );
      va.y -= ( double )( v1->y );
      va.z -= ( double )( v1->z );
      vb.x -= ( double )( v1->x );
      vb.y -= ( double )( v1->y );
      vb.z -= ( double )( v1->z );
      // cross (vc)
      CROSSPRODUCT( &va, &vb, &vc );
      // 2-norm
      double sq = 0;
      sq += ( vc.x * vc.x );
      sq += ( vc.y * vc.y );
      sq += ( vc.z * vc.z );
      double norm2 = sqrt( sq );
      if( norm2 == 0 )
      {
         norm2 = 0.0000005;
      }
      // Area
      double area = norm2 * 0.5;
      // Tri center
      DVertexS center;
      center.x = (( double )( v1->x + v2->x + v3->x )) / 3.0;
      center.y = (( double )( v1->y + v2->y + v3->y )) / 3.0;
      center.z = (( double )( v1->z + v2->z + v3->z )) / 3.0;
      // Normal vector
      Vertex nvec;
      nvec.x = vc.x / norm2;
      nvec.y = vc.y / norm2;
      nvec.z = vc.z / norm2;
      // Volume
      //totalvol += (( center.x + center.y + center.z ) * ( nvec.x + nvec.y + nvec.z ) * area );
      totalvol += (( center.x ) * ( nvec.x ) * area );
//printf( "center = %g %g %g\nvnorm  = %g %g %g\narea   = %g\n", center.x, center.y, center.z, nvec.x, nvec.y, nvec.z, area );
      totalarea += area;
   }
   /**/

   // Get volume using divergence theorem calculation
   /**
   double totalvolold = 0;
   for( List *ps = seltris; ps != NULL; ps = ps->next )
   {
      Surface *s = ( Surface * )( ps->data );
      Vertex *v1 = (( SVertex * )( s->vertlist->data ))->v;
      Vertex *v2 = (( SVertex * )( s->vertlist->next->data ))->v;
      Vertex *v3 = (( SVertex * )( s->vertlist->next->next->data ))->v;
      totalvolold += ((((( double )( v2->y )) - (( double )( v1->y ))) * ((( double )( v3->z )) - (( double )( v1->z )))) -
                   (((( double )( v2->z )) - (( double )( v1->z ))) * ((( double )( v3->y )) - (( double )( v1->y ))))) *
                  ((( double )( v1->x )) + (( double )( v2->x )) + (( double )( v3->x )));
   }
   totalvolold /= 6.0;
   printf( "New = %g\nOld = %g\n", totalvol, totalvolold );
   /**/

   // Free tri list
   list_free( &seltris );

   // Print results
   display_message( "calcvol : Total selected volume is %g", totalvol );
   if( complex_geometry && open_geometry )
   {
      message_dialog( "** WARNING** The selection is not an enclosed volume.  The resulting calculation below may be inaccurate.\n\nSome edges in the current selection are made by the intersection of 3 or more surfaces.  This may cause inaccurate volume calculation results.\n\nTotal volume of selection is %g", totalvol );
   }
   else if( complex_geometry )
   {
      message_dialog( "** WARNING** Some edges in the current selection are made by the intersection of 3 or more surfaces.  This may cause inaccurate volume calculation results.\n\nTotal volume of selection is %g", totalvol );
   }
   else if( open_geometry )
   {
      message_dialog( "** WARNING** The selection is not an enclosed volume.  The resulting calculation below may be inaccurate.\n\nTotal volume of selection is %g", totalvol );
   }
   else
   {
      message_dialog( "Total volume of selection is %g", totalvol );
   }

}
//----------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{
   static bool firstTime = true;
   if( firstTime )
   {
      ac_add_command( "ddh_calc_volume", ddh_calc_volume );
      ac_add_tools_menu_item( "Calculate selected volume",
                              "ac3d ddh_calc_volume",
                              "Calculate the volume of the selection" );
      firstTime = false;
   }
   return 0;
}
//----------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//----------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Calculate volume - version 1.0 - Dennis Hawthorne" );
}
//----------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Calculate volume - version 1.0 - Dennis Hawthorne - calculates the volume of the selection" );
}
//----------------------------------------------------------------------------------
